/*---------------------------------------------------------------------------*/

/*!  \file  analyzer.c
     \brief  -- XUDF (eXtened User Definable Format) lexical analyzer.  */

/*---------------------------------------------------------------------------*/

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tokenizer.h"
#include "analyzer.h"
#include "tree.h"
#include "grammer.h"

#define CHECK_TOKEN(fp,xt,t) \
{\
    if(xudf_get_token((fp),(xt)) == EOF) \
    { \
        return XUDF_ERROR; \
    } \
    if(xt->type != (t)) \
    { \
        return XUDF_ERROR; \
    } \
}

/*---------------------------------------------------------------------------*/
static int xudf_analyze_def(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot);
static int xudf_analyze_def_variable(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot);
static int xudf_analyze_def_variable_type(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot);
static int xudf_analyze_def_variable_type_class(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot);
static int xudf_analyze_def_variable_select(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_analyze_def_variable_unit(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_analyze_def_class(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot);
static int xudf_analyze_data(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot);
static int xudf_analyze_data_variable(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot);
static int xudf_analyze_data_variable_value(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot);
static int xudf_analyze_data_variable_value_array(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot);
static int xudf_analyze_data_variable_value_struct(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot);
static int xudf_analyze_header(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot);
static int xudf_analyze_record(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot);
static int xudf_analyze_unit(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_analyze_unit_constant(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_analyze_unit_unit(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_analyze_unit_unit_base(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_is_primitive_type(char *name);

/*---------------------------------------------------------------------------*/
int xudf_analyze(FILE *fp,FILE *fp2)
{
    xudf_token *xt;
    xudf_def_node *defroot,*headerroot,*classroot;

    fprintf(fp2,FILE_HEADER);
    fprintf(fp2,"\n");
    fprintf(fp2,HASH RECORD " " ASSIGNOR " " HASH_TOP_LEFT HASH_TOP_RIGHT TERMINATOR "\n");
    fprintf(fp2,"\n");

    xt = xudf_create_token();
    defroot = xudf_new_def_tree("def");
    headerroot = xudf_new_def_tree("header");
    classroot = xudf_new_def_tree("class");

    while(xudf_get_token(fp,xt) != EOF)
    {
        if(strcmp(xt->string,"def") == 0)
        {
/*             fprintf(fp2,COMMENT_BEGIN(DEF) "\n"); */
            xudf_analyze_def(fp,fp2,xt,defroot,classroot);
/*             fprintf(fp2,COMMENT_END(DEF) "\n"); */
/*             fprintf(fp2,"\n"); */

/*             xudf_print_def_tree(fp2,defroot); */
/*             xudf_print_def_tree(fp2,classroot); */
        }
        else if(strcmp(xt->string,"global") == 0)
        {
/*             fprintf(fp2,COMMENT_BEGIN(GLOBAL) "\n"); */
            xudf_analyze_def(fp,fp2,xt,defroot,classroot);
/*             fprintf(fp2,COMMENT_END(GLOBAL) "\n"); */
/*             fprintf(fp2,"\n"); */
        }
        else if(strcmp(xt->string,"data") == 0)
        {
            fprintf(fp2,HASH DATA " " ASSIGNOR " " HASH_TOP_LEFT "\n");
            xudf_analyze_data(fp,fp2,xt,defroot);
            fprintf(fp2,HASH_TOP_RIGHT TERMINATOR "\n");
            fprintf(fp2,"\n");
        }
        else if(strcmp(xt->string,"header") == 0)
        {
            fprintf(fp2,HASH HEADER " " ASSIGNOR " " HASH_TOP_LEFT "\n");
            xudf_analyze_header(fp,fp2,xt,headerroot,classroot);
            fprintf(fp2,HASH_TOP_RIGHT TERMINATOR "\n");
            fprintf(fp2,"\n");
        }
        else if(strcmp(xt->string,"record") == 0)
        {
            fprintf(fp2,VARIABLE RECORD);
            xudf_analyze_record(fp,fp2,xt,defroot);
            fprintf(fp2,HASH_RIGHT TERMINATOR "\n");
            fprintf(fp2,"\n");
        }
        else if(strcmp(xt->string,"unit") == 0)
        {
/*             fprintf(fp2,COMMENT_BEGIN(UNIT) "\n"); */
            xudf_analyze_unit(fp,fp2,xt);
/*             fprintf(fp2,COMMENT_END(UNIT) "\n"); */
/*             fprintf(fp2,"\n"); */
        }
        else
        {
            fprintf(stderr,"warning: obsolete style comment was found.\n");
        }
    }

/*     xudf_print_def_tree(fp2,defroot); */
/*     xudf_print_def_tree(fp2,headerroot); */

    xudf_free_def_tree(defroot);
    xudf_free_def_tree(headerroot);
    xudf_free_def_tree(classroot);
    xudf_destroy_token(xt);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot)
{
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
                return TRUE;

            case TT_LEFT_BRACKET:
                xudf_analyze_def_variable_unit(fp,fp2,xt);
                break;

            case TT_SPECIFIER:
                if(strcmp(xt->string,"class") == 0)
                {
                    xudf_analyze_def_class(fp,fp2,xt,defroot,classroot);
                }
                else
                {
                    xudf_analyze_def_variable(fp,fp2,xt,defroot,classroot);
                }
                break;

            case TT_STRING:
                fprintf(stderr,"warning: obsolete style hint was found.\n");
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def_class(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot)
{
    xudf_def_node *classcur;
    int i;

    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;

    classcur = xudf_add_child_def_node(classroot,xt->string);
/*     fprintf(fp2,"%s : class\n",xt->string); */

    CHECK_TOKEN(fp,xt,TT_COLON);

    i = xudf_analyze_def(fp,fp2,xt,classcur,classroot);

/*     fprintf(fp2,"end of class\n"); */

    return i;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def_variable(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot)
{
    int al;
    xudf_def_node *defcur;

    al = 0;

    defcur = xudf_add_child_def_node(defroot,xt->string);
/*     fprintf(fp2,"%s : ",xt->string); */

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_COLON:
                return xudf_analyze_def_variable_type(fp,fp2,xt,defcur,classroot);

            case TT_LEFT_BRACKET:
                CHECK_TOKEN(fp,xt,TT_RIGHT_BRACKET);
                al++;
/*                 fprintf(fp2,"array of "); */
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def_variable_unit(FILE *fp,FILE *fp2,xudf_token *xt)
{
/*     fprintf(fp2,"unit : [ "); */
    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACKET:
/*                 fprintf(fp2,"]\n"); */
                return TRUE;

            default:
/*                 fprintf(fp2,"%s ",xt->string); */
                break;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def_variable_type(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot)
{
    int i;

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_LEFT_BRACE:
                ungetc((int)' ',fp);
                ungetc((int)'{',fp);
                ungetc((int)' ',fp);
/*                 fprintf(fp2,"struct\n"); */
                i = xudf_analyze_def(fp,fp2,xt,defroot,classroot);
/*                 fprintf(fp2,"end of struct\n"); */
                return i;

            case TT_SPECIFIER:
                if(strcmp(xt->string,"select") == 0)
                {
/*                     fprintf(fp2,"select\n"); */
                    return xudf_analyze_def_variable_select(fp,fp2,xt);
                }
                else
                {
                    if(xudf_is_primitive_type(xt->string))
                    {
/*                         fprintf(fp2,"%s\n",xt->string); */
                    }
                    else
                    {
/*                         fprintf(fp2,"class %s\n",xt->string); */
                        return xudf_analyze_def_variable_type_class(fp,fp2,xt,defroot,classroot);
                    }
                    return TRUE;
                }

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def_variable_type_class(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot)
{
    xudf_def_node *classcur;

    classcur = xudf_find_def_node_by_name_or_exit(classroot,xt->string);
    xudf_copy_child_def_node(defroot,classcur);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_def_variable_select(FILE *fp,FILE *fp2,xudf_token *xt)
{
/*     fprintf(fp2,"select : "); */
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
/*                 fprintf(fp2,"\n"); */
                return TRUE;

            case TT_STRING:
/*                 fprintf(fp2,"%s ",xt->string); */
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_data(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot)
{
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
                return TRUE;

            case TT_SPECIFIER:
                xudf_analyze_data_variable(fp,fp2,xt,defroot);
                break;

            default:
                return XUDF_ERROR;
                break;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_data_variable(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot)
{
    xudf_def_node *defcur;
    int al;
    static char name_buffer[XUDF_TOKEN_SIZE];

    al = 0;

    defcur = xudf_find_def_node_by_name_or_exit(defroot,xt->string);

    strcpy(name_buffer,xt->string);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_COLON:
                fprintf(fp2,KEY("%s") " " ASSOCIATOR " ",name_buffer);
                return xudf_analyze_data_variable_value(fp,fp2,xt,defcur);

            case TT_LEFT_BRACKET:
                CHECK_TOKEN(fp,xt,TT_RIGHT_BRACKET);
                al++;
                break;

            case TT_SPECIFIER:
                fprintf(stderr,"warning: obsolete style comment was found.\n");
                strcpy(name_buffer,xt->string);
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_data_variable_value(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot)
{
    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_LEFT_BRACKET:
                return xudf_analyze_data_variable_value_array(fp,fp2,xt,defroot);

            case TT_LEFT_BRACE:
                return xudf_analyze_data_variable_value_struct(fp,fp2,xt,defroot);

            case TT_STRING:
            case TT_INTEGER:
            case TT_FLOAT:
                fprintf(fp2,"%s" SEPARATOR "\n",xt->string);
                return TRUE;

            case TT_SPECIFIER:
                fprintf(stderr,"warning: obsolete style comment was found.\n");
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_data_variable_value_array(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot)
{
    fprintf(fp2,ARRAY_LEFT " ");

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACKET:
                fprintf(fp2,ARRAY_RIGHT SEPARATOR "\n");
                return TRUE;

            case TT_LEFT_BRACKET:
                xudf_analyze_data_variable_value_array(fp,fp2,xt,defroot);
                break;

            case TT_LEFT_BRACE:
                xudf_analyze_data_variable_value_struct(fp,fp2,xt,defroot);
                break;

            case TT_STRING:
            case TT_INTEGER:
            case TT_FLOAT:
                fprintf(fp2,"%s" SEPARATOR " ",xt->string);
                break;

            case TT_SPECIFIER:
                fprintf(stderr,"warning: obsolete style comment was found.\n");
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_data_variable_value_struct(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot)
{
    int i;
    xudf_def_node *defcur;

    fprintf(fp2,HASH_LEFT " ");

    i = 0;
    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
                fprintf(fp2,HASH_RIGHT SEPARATOR "\n");
                return TRUE;

            case TT_LEFT_BRACKET:
                defcur = xudf_find_def_node_by_index_or_exit(defroot,i++);
                fprintf(fp2,KEY("%s") " " ASSOCIATOR " ",defcur->name);
                xudf_analyze_data_variable_value_array(fp,fp2,xt,defcur);
                break;

            case TT_LEFT_BRACE:
                defcur = xudf_find_def_node_by_index_or_exit(defroot,i++);
                fprintf(fp2,KEY("%s") " " ASSOCIATOR " ",defcur->name);
                xudf_analyze_data_variable_value_struct(fp,fp2,xt,defcur);
                break;

            case TT_STRING:
            case TT_INTEGER:
            case TT_FLOAT:
                defcur = xudf_find_def_node_by_index_or_exit(defroot,i++);
                fprintf(fp2,KEY("%s") " " ASSOCIATOR " ",defcur->name);
                fprintf(fp2,"%s" SEPARATOR " ",xt->string);
                break;

            case TT_SPECIFIER:
                fprintf(stderr,"warning: obsolete style comment was found.\n");
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_header(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot,xudf_def_node *classroot)
{
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
                return TRUE;

            case TT_SPECIFIER:
                if(strcmp(xt->string,"def") == 0)
                {
/*                     fprintf(fp2,COMMENT_BEGIN(DEF) "\n"); */
                    xudf_analyze_def(fp,fp2,xt,defroot,classroot);
/*                     fprintf(fp2,COMMENT_END(DEF) "\n"); */
/*                     fprintf(fp2,"\n"); */
                    break;
                }
                else if(strcmp(xt->string,"data") == 0)
                {
                    fprintf(fp2,KEY(DATA) " " ASSOCIATOR " " HASH_LEFT "\n");
                    xudf_analyze_data(fp,fp2,xt,defroot);
                    fprintf(fp2,HASH_RIGHT " " SEPARATOR "\n");
                    fprintf(fp2,"\n");
                    break;
                }
                else
                    return XUDF_ERROR;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_record(FILE *fp,FILE *fp2,xudf_token *xt,xudf_def_node *defroot)
{
    CHECK_TOKEN(fp,xt,TT_STRING);
    fprintf(fp2,ACCESSOR_LEFT "%s" ACCESSOR_RIGHT " " ASSIGNOR " " HASH_LEFT "\n",xt->string);
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
                return TRUE;

            case TT_SPECIFIER:
                if(strcmp(xt->string,"data") == 0)
                {
                    fprintf(fp2,KEY(DATA) " " ASSOCIATOR " " HASH_LEFT "\n");
                    xudf_analyze_data(fp,fp2,xt,defroot);
                    fprintf(fp2,HASH_RIGHT " " SEPARATOR "\n");
                    fprintf(fp2,"\n");
                    break;
                }
                else
                    return XUDF_ERROR;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_unit(FILE *fp,FILE *fp2,xudf_token *xt)
{
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACE:
                return TRUE;

            case TT_LEFT_BRACKET:
                xudf_analyze_unit_unit(fp,fp2,xt);
                break;

            case TT_SPECIFIER:
                xudf_analyze_unit_constant(fp,fp2,xt);
                break;

            default:
                return XUDF_ERROR;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_unit_constant(FILE *fp,FILE *fp2,xudf_token *xt)
{
/*     fprintf(fp2,"constant : %s = ",xt->string); */

    CHECK_TOKEN(fp,xt,TT_EQUAL);

    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;

/*     fprintf(fp2,"%s\n",xt->string); */

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_unit_unit(FILE *fp,FILE *fp2,xudf_token *xt)
{
    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;
/*     fprintf(fp2,"unit : [ %s ] = ",xt->string); */

    CHECK_TOKEN(fp,xt,TT_RIGHT_BRACKET);
    CHECK_TOKEN(fp,xt,TT_EQUAL);

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_LEFT_BRACKET:
                return xudf_analyze_unit_unit_base(fp,fp2,xt);

            default:
/*                 fprintf(fp2,"%s ",xt->string); */
                break;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_analyze_unit_unit_base(FILE *fp,FILE *fp2,xudf_token *xt)
{
/*     fprintf(fp2,"[ "); */

    while(xudf_get_token(fp,xt) != EOF)
    {
        switch(xt->type)
        {
            case TT_RIGHT_BRACKET:
/*                 fprintf(fp2,"]\n"); */
                return TRUE;

            default:
/*                 fprintf(fp2,"%s ",xt->string); */
                break;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_is_primitive_type(char *name)
{
    static char *primitives[] = {"int","long","short","float",
                                 "double","single","string","select",
                                 "KEY","ID"};
    static int number_of_primitives = 10;
    int i;

    for(i = 0;i < number_of_primitives;i++)
    {
        if(strcmp(name,primitives[i]) == 0)
            return TRUE;
    }

    return FALSE;
}

/*---------------------------------------------------------------------------*/
