/*---------------------------------------------------------------------------*/

/*!  \file  tree.c
     \brief  XUDF (eXtened User Definable Format) data structure tree.  */

/*---------------------------------------------------------------------------*/

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xmalloc.h"
#include "tree.h"

/*---------------------------------------------------------------------------*/
static void xudf_print_def_tree_recursive(FILE *fp,xudf_def_node *xdef,int n);

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_new_def_tree(char *xudfname)
{
    xudf_def_node *xdef;

    xdef = xcalloc(1,sizeof(xudf_def_node));
    xdef->name = xstrdup(xudfname);
    xdef->next = NULL;
    xdef->child = NULL;

    return xdef;
}

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_add_next_def_node(xudf_def_node *xdef,
                                      char *xdefname)
{
    xudf_def_node *xdeflast;

    xdeflast = xdef;
    while(xdeflast->next != NULL)
    {
        xdeflast = xdeflast->next;
    }

    xdeflast->next = xcalloc(1,sizeof(xudf_def_node));
    xdeflast->next->name = xstrdup(xdefname);
    xdeflast->next->next = NULL;
    xdeflast->next->child = NULL;

    return xdeflast->next;
}

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_add_child_def_node(xudf_def_node *xdef,
                                       char *xdefname)
{
    xudf_def_node *xdeflast;

    if(xdef->child != NULL)
    {
        xdeflast = xdef->child;
        while(xdeflast->next != NULL)
        {
            xdeflast = xdeflast->next;
        }

        xdeflast->next = xcalloc(1,sizeof(xudf_def_node));
        xdeflast->next->name = xstrdup(xdefname);
        xdeflast->next->next = NULL;
        xdeflast->next->child = NULL;

        return xdeflast->next;
    }
    else
    {
        xdeflast = xdef;

        xdeflast->child = xcalloc(1,sizeof(xudf_def_node));
        xdeflast->child->name = xstrdup(xdefname);
        xdeflast->child->next = NULL;
        xdeflast->child->child = NULL;

        return xdeflast->child;
    }

}

/*---------------------------------------------------------------------------*/
void xudf_copy_child_def_node(xudf_def_node *xdef,xudf_def_node *xsub)
{
    xudf_def_node *xcur,*xchild;

    if(xsub == NULL)
        return;

    xcur = xsub->child;

    while(xcur != NULL)
    {
        xchild = xudf_add_child_def_node(xdef,xcur->name);
        xudf_copy_child_def_node(xchild,xcur);
        xcur = xcur->next;
    }
}

/*---------------------------------------------------------------------------*/
void xudf_free_def_tree(xudf_def_node *xdef)
{
    if(xdef != NULL)
    {
        xudf_free_def_tree(xdef->child);
        xudf_free_def_tree(xdef->next);

        if(xdef->name != NULL)
            xfree(xdef->name);
        xfree(xdef);
    }
}

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_find_def_node_by_name(xudf_def_node *xdef,char *xdefname)
{
    if(xdef == NULL)
        return NULL;

    xdef = xdef->child;

    while(xdef != NULL)
    {
        if(strcmp(xdef->name,xdefname) == 0)
        {
            return xdef;
        }
        xdef = xdef->next;
    }

    return NULL;
}

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_find_def_node_by_name_or_exit(xudf_def_node *xdef,char *xdefname)
{
    xudf_def_node *xdeffound;

    xdeffound = xudf_find_def_node_by_name(xdef,xdefname);

    if(xdeffound != NULL)
        return xdeffound;

    fprintf(stderr,"error: cannot find a definition for `%s' in `%s'\n",
            xdefname,xdef->name);
    exit(EXIT_FAILURE);
}

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_find_def_node_by_index(xudf_def_node *xdef,int index)
{
    int i;

    if((xdef == NULL) || (index < 0))
        return NULL;

    xdef = xdef->child;
    for(i = 0;i < index;i++)
    {
        xdef = xdef->next;
        if(xdef == NULL)
            return NULL;
    }

    return xdef;
}

/*---------------------------------------------------------------------------*/
xudf_def_node *xudf_find_def_node_by_index_or_exit(xudf_def_node *xdef,int index)
{
    xudf_def_node *xdeffound;

    xdeffound = xudf_find_def_node_by_index(xdef,index);

    if(xdeffound != NULL)
        return xdeffound;

    fprintf(stderr,"error: cannot find a definition for index %d in `%s'\n",
            index,xdef->name);
    exit(EXIT_FAILURE);
}

/*---------------------------------------------------------------------------*/
void xudf_print_def_tree(FILE *fp,xudf_def_node *xdef)
{
    fprintf(fp,"--[[tree\n");
    xudf_print_def_tree_recursive(fp,xdef,0);
    fprintf(fp,"--tree]]\n");
}

/*---------------------------------------------------------------------------*/
void xudf_print_def_node(FILE *fp,xudf_def_node *xdef)
{
    fprintf(fp,"--[[tree\n");
    if(xdef != NULL)
    {
        fprintf(fp,"%s\n",xdef->name);

        xudf_print_def_tree_recursive(fp,xdef->child,1);
    }
    fprintf(fp,"--tree]]\n");
}

/*---------------------------------------------------------------------------*/
void xudf_print_def_tree_recursive(FILE *fp,xudf_def_node *xdef,int n)
{
    int i;

    if(xdef != NULL)
    {
        for(i = 0;i < (n - 1);i++)
            printf("  ");
        for(;i < n;i++)
            printf("+-");

        fprintf(fp,"%s\n",xdef->name);

        xudf_print_def_tree_recursive(fp,xdef->child,n + 1);
        xudf_print_def_tree_recursive(fp,xdef->next,n);
    }
}

/*---------------------------------------------------------------------------*/
