/*---------------------------------------------------------------------------*/

/*!  \file  analyzer.c
     \brief  XUDF (eXtened User Definable Format) lexical analyzer.  */

/*---------------------------------------------------------------------------*/

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdio.h>
#include <string.h>
#include "tokenizer.h"
#include "analyzer.h"

#define CHECK_TOKEN(fp,xt,t) \
{\
    if(xudf_get_token((fp),(xt)) == EOF) \
    { \
        return XUDF_ERROR; \
    } \
    if(xt->type != (t)) \
    { \
        return XUDF_ERROR; \
    } \
}

/*---------------------------------------------------------------------------*/
static int xudf_process_backslash(FILE *fp,FILE *fp2,xudf_token *xt,
                                  xudf_path_list *xlist,
                                  xudf_definition_alist *xalist);
static int xudf_process_begin(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_process_end(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_process_include(FILE *fp,FILE *fp2,xudf_token *xt,
                                xudf_path_list *xlist,
                                xudf_definition_alist *xalist);
static int xudf_process_less_than(FILE *fp,FILE *fp2,xudf_token *xt);
static int xudf_process_define(FILE *fp,FILE *fp2,xudf_token *xt,xudf_definition_alist *xalist);
static int xudf_process_definition(FILE *fp,FILE *fp2,xudf_token *xt,xudf_definition_alist *xalist);

/*---------------------------------------------------------------------------*/
int xudf_process_macro(FILE *fp,FILE *fp2,xudf_path_list *xlist,
                       xudf_definition_alist *xalist)
{
    int i,n;
    xudf_token *xt;

    n = 1;
    xt = xudf_create_token();

    while(xudf_get_token(fp,xt) != EOF)
    {
        for(i = 0;i < xt->newlines;i++)
        {
            fprintf(fp2,"\n");
            n++;
        }

        if(xt->type == TT_BACKSLASH)
        {
            if(xudf_process_backslash(fp,fp2,xt,xlist,xalist) == XUDF_ERROR)
                return XUDF_ERROR;
        }
        else if(xt->type == TT_LESS_THAN)
        {
            if(xudf_process_less_than(fp,fp2,xt) == XUDF_ERROR)
                return XUDF_ERROR;
        }
        else
        {
            fprintf(fp2,"%s ",xt->string);
        }
    }
    fprintf(fp2,"\n");
    xudf_destroy_token(xt);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
static int xudf_process_backslash(FILE *fp,FILE *fp2,xudf_token *xt,
                                  xudf_path_list *xlist,
                                  xudf_definition_alist *xalist)
{
    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;

    if(xt->type == TT_BEGIN)
    {
        return xudf_process_begin(fp,fp2,xt);
    }
    else if(xt->type == TT_END)
    {
        return xudf_process_end(fp,fp2,xt);
    }
    else if(xt->type == TT_INCLUDE)
    {
        return xudf_process_include(fp,fp2,xt,xlist,xalist);
    }
    else if(xt->type == TT_DEFINE)
    {
        return xudf_process_define(fp,fp2,xt,xalist);
    }
    else if(xt->type == TT_SPECIFIER)
    {
        return xudf_process_definition(fp,fp2,xt,xalist);
    }
    else
    {
        return XUDF_ERROR;
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*/
static int xudf_process_begin(FILE *fp,FILE *fp2,xudf_token *xt)
{
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;
    if(xt->type == TT_RECORD)
    {
        fprintf(fp2,"%s ",xt->string);

        CHECK_TOKEN(fp,xt,TT_RIGHT_BRACE);

        CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

        if(xudf_get_token(fp,xt) == EOF)
            return XUDF_ERROR;
        fprintf(fp2,"%s {",xt->string);
    }
    else if(xt->type == TT_GLOBAL_DEF)
    {
        fprintf(fp2,"global {");
    }
    else
    {
        fprintf(fp2,"%s { ",xt->string);
    }

    CHECK_TOKEN(fp,xt,TT_RIGHT_BRACE);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
static int xudf_process_end(FILE *fp,FILE *fp2,xudf_token *xt)
{
    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;
    fprintf(fp2,"} ");

    CHECK_TOKEN(fp,xt,TT_RIGHT_BRACE);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
static int xudf_process_include(FILE *fp,FILE *fp2,xudf_token *xt,
                                xudf_path_list *xlist,
                                xudf_definition_alist *xalist)
{
    char *s;
    FILE *fp3;
    char *filename;

    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);

    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;

    s = xt->string + 1;
    while(*s != '"')
        s++;
    *s = '\0';
    s = xt->string + 1;

    filename = xudf_find_file_from_path_list_or_exit(xlist,s);

    if(filename != NULL)
    {
            /*  include file  */
        if((fp3 = fopen(filename,"rt")) == NULL)
        {
            return XUDF_ERROR;
        }

        fprintf(fp2,"\n\n");
        xudf_process_macro(fp3,fp2,xlist,xalist);
        fclose(fp3);
    }
    else
    {
        fprintf(stderr,"error: cannot find file '%s'\n",s);
        return XUDF_ERROR;
    }

    CHECK_TOKEN(fp,xt,TT_RIGHT_BRACE);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_process_less_than(FILE *fp,FILE *fp2,xudf_token *xt)
{
    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;

    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;
    fprintf(fp2,"%s ",xt->string);

    CHECK_TOKEN(fp,xt,TT_GREATER_THAN);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_process_define(FILE *fp,FILE *fp2,xudf_token *xt,xudf_definition_alist *xalist)
{
    int i;
    static char name_buffer[XUDF_DEFINITION_SIZE];
    static char definition_buffer[XUDF_DEFINITION_SIZE];

    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);
    if(xudf_get_token(fp,xt) == EOF)
        return XUDF_ERROR;
    strcpy(name_buffer,xt->string);
    CHECK_TOKEN(fp,xt,TT_RIGHT_BRACE);

    CHECK_TOKEN(fp,xt,TT_LEFT_BRACE);
    i = 1;
    definition_buffer[0] = '\0';
    while(TRUE)
    {
        if(xudf_get_token(fp,xt) == EOF)
            return XUDF_ERROR;

        if(xt->type == TT_LEFT_BRACE)
        {
            i++;
        }
        else if(xt->type == TT_RIGHT_BRACE)
        {
            i--;
            if(!(i > 0))
                break;
        }

        if(definition_buffer[0] != '\0')
            strcat(definition_buffer," ");
        strcat(definition_buffer,xt->string);
    }

    xudf_add_definition_alist(xalist,name_buffer,definition_buffer);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
int xudf_process_definition(FILE *fp,FILE *fp2,xudf_token *xt,xudf_definition_alist *xalist)
{
    char *s;

    s = xudf_find_definition_from_definition_alist_or_exit(xalist,xt->string);
    if(s == NULL)
        return XUDF_ERROR;

    fprintf(fp2,"%s ",s);

    return TRUE;
}

/*---------------------------------------------------------------------------*/
