/*---------------------------------------------------------------------------*/

/*!  \file  tokenizer.c
     \brief  XUDF (eXtened User Definable Format) tokenizer.  */

/*---------------------------------------------------------------------------*/

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "common.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "xmalloc.h"

#include "tokenizer.h"

/*---------------------------------------------------------------------------*/
static int xudf_char_is_white_space(char c);
static int xudf_skip_spaces_and_comments(FILE *fp,xudf_token *xt);
static int xudf_skip_c_comment(FILE *fp,xudf_token *xt);
static int xudf_skip_cpp_comment(FILE *fp,xudf_token *xt);
static void xudf_analyze_token_type(xudf_token *xt);
static int xudf_char_is_alnumul(char c);
static int xudf_char_is_alphaul(char c);
static int xudf_char_is_alnumul(char c);
static int xudf_char_is_alphaul(char c);

/*---------------------------------------------------------------------------*/
xudf_token *xudf_create_token(void)
{
    xudf_token *xt;

    xt = xcalloc(1,sizeof(xudf_token));
    if(xt != NULL)
        xt->string = xcalloc(XUDF_TOKEN_SIZE,sizeof(char));

    return xt;
}

/*---------------------------------------------------------------------------*/
void xudf_destroy_token(xudf_token *xt)
{
    if(xt != NULL)
    {
        if(xt->string != NULL)
            xfree(xt->string);
        xfree(xt);
    }
}

/*---------------------------------------------------------------------------*/
int xudf_get_token(FILE *fp,xudf_token *xt)
{
    int i;
    char c;
    char *s;

    s = xt->string;
    xt->newlines = 0;
    xudf_skip_spaces_and_comments(fp,xt);

    if((i = fgetc(fp)) == EOF)
        goto err;

    c = (char)i;
    if((c == '\\') ||
       (c == '{') ||
       (c == '}') ||
       (c == '[') ||
       (c == ']') ||
       (c == ':') ||
       (c == '<') ||
       (c == '>'))
    {
        *s++ = c;
        goto ret;
    }
    else if(c == ';')
    {
        goto ign;
    }
    else if(c == ',')
    {
        goto ign;
    }
    else if(c == '"')
    {
        *s++ = c;
        while((i = fgetc(fp)) != EOF)
        {
            c = (char)i;
            if(c == '\\')
            {
                *s++ = c;
                if((i = fgetc(fp)) == EOF)
                    goto err;
                *s++ = (char)i;
            }
            else if(c == '"')
            {
                *s++ = c;
                goto ret;
            }
            else
            {
                *s++ = c;
            }
        }
        goto err;
    }
    else if(c == '\'')
    {
        *s++ = c;
        while((i = fgetc(fp)) != EOF)
        {
            c = (char)i;
            if(c == '\\')
            {
                *s++ = c;
                if((i = fgetc(fp)) == EOF)
                    goto err;
                *s++ = (char)i;
            }
            else if(c == '\'')
            {
                *s++ = c;
                goto ret;
            }
            else
            {
                *s++ = c;
            }
        }
        goto err;
    }
    else
    {
        *s++ = c;
        while((i = fgetc(fp)) != EOF)
        {
            c = (char)i;
            if((xudf_char_is_white_space(c)) ||
               (c == ';') ||
               (c == '{') ||
               (c == '}') ||
               (c == '[') ||
               (c == ']') ||
               (c == ':') ||
               (c == ',') ||
               (c == '/') ||
               (c == '"') ||
               (c == '<') ||
               (c == '>') ||
               (c == '\''))
            {
                ungetc(c,fp);
                goto ret;
            }
            else
                *s++ = c;
        }
        goto err;
    }

  err:
    xt->type = TT_UNKNOWN;
    xt->string[0] = '\0';
    return EOF;

  ret:
    *s = '\0';
    xudf_analyze_token_type(xt);
    return (int)(s - xt->string);

  ign:
    return xudf_get_token(fp,xt);
}

/*---------------------------------------------------------------------------*/
static int xudf_char_is_white_space(char c)
{
    return ((c == ' ') ||
            (c == '\n') ||
            (c == '\t') ||
            (c == 0x0a) ||
            (c == 0x0d));
}

/*---------------------------------------------------------------------------*/
static int xudf_skip_spaces_and_comments(FILE *fp,xudf_token *xt)
{
    int i;
    char c;

        /*  Skip white space and C/C++ style comment.  */
    while((i = fgetc(fp)) != EOF)
    {
        c = (char)i;

            /*  Skip white space.  */
        if(!xudf_char_is_white_space(c))
        {
            if(c == '/')
            {
                if((i = fgetc(fp)) == EOF)
                {
                    ungetc((int)'/',fp);
                    return EOF;
                }
                c = (char)i;

                if(c == '*')
                {
                        /*  Skip C style comment.  */
                    if(xudf_skip_c_comment(fp,xt) == EOF)
                        return EOF;
                }
                else if(c == '/')
                {
                        /*  Skip C++ style comment.  */
                    if(xudf_skip_cpp_comment(fp,xt) == EOF)
                        return EOF;
                }
                else
                {
                    ungetc(i,fp);
                    ungetc((int)'/',fp);
                    return 0;
                }
            }
            else
            {
                ungetc(i,fp);
                return 0;
            }
        }
        else if(c == '\n')
        {
            xt->newlines++;
        }
    }

    return EOF;
}

/*---------------------------------------------------------------------------*/
static int xudf_skip_c_comment(FILE *fp,xudf_token *xt)
{
    int i;
    char c;

        /*  Skip C style comment.  */
    while((i = fgetc(fp)) != EOF)
    {
        c = (char)i;

        if(c == '*')
        {
            if((i = fgetc(fp)) == EOF)
            {
                return EOF;
            }
            c = (char)i;

            if(c == '/')
                return 0;
            else if(c == '\n')
                xt->newlines++;

            ungetc(i,fp);
        }
    }

    return EOF;
}

/*---------------------------------------------------------------------------*/
static int xudf_skip_cpp_comment(FILE *fp,xudf_token *xt)
{
    int i;
    char c;

        /*  Skip C++ style comment.  */
    while((i = fgetc(fp)) != EOF)
    {
        c = (char)i;

        if(c == '\n')
        {
            xt->newlines++;
            return 0;
        }
    }

    return EOF;
}

/*---------------------------------------------------------------------------*/
static void xudf_analyze_token_type(xudf_token *xt)
{
    int i,n;

    n = strlen(xt->string);

    if(n == 1)
    {
        switch(xt->string[0])
        {
            case '\\':
                xt->type = TT_BACKSLASH;
                break;
            case '{':
                xt->type = TT_LEFT_BRACE;
                break;
            case '}':
                xt->type = TT_RIGHT_BRACE;
                break;
            case '[':
                xt->type = TT_LEFT_BRACKET;
                break;
            case ']':
                xt->type = TT_RIGHT_BRACKET;
                break;
            case ':':
                xt->type = TT_COLON;
                break;
            case '>':
                xt->type = TT_GREATER_THAN;
                break;
            case '<':
                xt->type = TT_LESS_THAN;
                break;
            default:
                xt->type = TT_OTHER;
                break;
        }

        if(xt->type != TT_OTHER)
            return;
    }

    if(xt->string[0] == '"')
        xt->type = TT_STRING;
    else if(xt->string[0] == '\'')
    {
        xt->type = TT_STRING;
        xt->string[0] = '"';
        xt->string[strlen(xt->string) - 1] = '"';
    }
    else if(strcmp(xt->string,"begin") == 0)
        xt->type = TT_BEGIN;
    else if(strcmp(xt->string,"end") == 0)
        xt->type = TT_END;
    else if(strcmp(xt->string,"header") == 0)
        xt->type = TT_HEADER;
    else if(strcmp(xt->string,"def") == 0)
        xt->type = TT_DEF;
    else if(strcmp(xt->string,"data") == 0)
        xt->type = TT_DATA;
    else if(strcmp(xt->string,"unit") == 0)
        xt->type = TT_UNIT;
    else if(strcmp(xt->string,"global_def") == 0)
        xt->type = TT_GLOBAL_DEF;
    else if(strcmp(xt->string,"include") == 0)
        xt->type = TT_INCLUDE;
    else if(strcmp(xt->string,"record") == 0)
        xt->type = TT_RECORD;
    else if(strcmp(xt->string,"define") == 0)
        xt->type = TT_DEFINE;
    else if(xudf_char_is_alphaul(xt->string[0]))
    {
        xt->type = TT_SPECIFIER;
        for(i = 1;i < n;i++)
        {
            if(!xudf_char_is_alnumul(xt->string[i]))
                xt->type = TT_UNKNOWN;
        }
    }
    else
        xt->type = TT_OTHER;
}

/*---------------------------------------------------------------------------*/
int xudf_char_is_alnumul(char c)
{
    return (isalnum(c) ||
            (c == '_'));
}

/*---------------------------------------------------------------------------*/
int xudf_char_is_alphaul(char c)
{
    return (isalpha(c) ||
            (c == '_'));
}
/*---------------------------------------------------------------------------*/
